#
# CVSWrapper
#
# A python wrapper class for the cvs command line tool
#
# Copyright (c) 2001 by beehive elektronische medien GmbH
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public License
# (MPL)
#

"""
$Id: CVSWrapper.py,v 1.9 2003/02/16 16:07:18 philipp Exp $
"""

import sys, os, re
from IRevisionController import IRevisionController
from ShellCommandProcessor import ShellCommandProcessor
from scramble import scramble

# unused yet maybe useful later
_cvs_states = {
    'Up-to-date':       'OK',
    'Locally Modified': 'LM',
    'Locally Added':    'LA',
    'Locally Removed':  'LR',
    'Needs Checkout':   'NC',
    'Needs Patch':      'NP',
    'Unknown':          'UN'
    }

_update_flags = {
    'U': 'Updated',
    'P': 'Patched',
    'A': 'To be added',
    'R': 'To be removed',
    'M': 'Locally modified',
    'C': 'Conflict',
    '?': 'Unknown'
    }

_import_flags = {
    'U': 'File already existed',
    'N': 'New file added',
    'C': 'Locally modified',
    'I': 'Ignored',
    'L': 'Symbolic link'
    }

_tag_flags = {
    'T': 'Tagged'
    }

_cvs_commands = ("init", "status", "checkout", "commit", "add", "remove",
                 "update", "import", "tag", "log", "diff")

class CVSError(Exception):
    pass

class CVSWrapper:
    """
    Wrapper class for the cvs command. Makes nearly all cvs functionality
    available as instance methods.
    """

    __implements__ = IRevisionController

    def __init__(self, cvs_bin="cvs", access_method="local", repository_path="/var/cvs/",
                 module="", work_dir="", server_addr="localhost", user_name="",
                 passwd="", cvs_rsh="rsh", comp_level="3", cvs_passfile="", debug=0):
        self.cvs_bin         = cvs_bin
        self.access_method   = access_method
        self.repository_path = repository_path
        self.module          = module
        self.work_dir        = work_dir
        self.server_addr     = server_addr
        self.user_name       = user_name
        self.passwd          = passwd
        self.cvs_rsh         = cvs_rsh
        self.comp_level      = comp_level
        self.cvs_passfile    = cvs_passfile
        self.debug           = debug
        
        # assemble repository string
        repository = ":%s:" % self.access_method
        if (self.access_method == "pserver" or self.access_method == "ext") and self.server_addr != "":
            if self.user_name != "":
                repository = repository + "%s@" % self.user_name
            repository = repository + "%s:" % self.server_addr
        repository = repository + self.repository_path
        self.repository = repository
        
        # generate environment dict for later use
        environment = {}
        if self.access_method == "pserver":
            environment['CVS_PASSFILE'] = self.cvs_passfile
        if self.access_method == "ext" and self.cvs_rsh != "":
            environment['CVS_RSH'] = self.cvs_rsh
        self.environment = environment

        if access_method == "pserver":
            self._login(passwd)

    def __del__(self):
        if self.access_method == "pserver":
            self._logout()

    def _login(self, passwd):
        """
        perform a CVS login
        does not really login but writes the correspondent line to a .cvspass-like
        file
        """
        cvspassline = self.repository + " " + scramble(passwd) + "\n"
        f = open(self.cvs_passfile, "a")
        f.write(cvspassline)
        f.close()

    def _logout(self):
        """
        perform a CVS logout - behaves just like login
        """
        lines = open(self.cvs_passfile, "r").readlines()
        newlines = filter(lambda l, r=self.repository: l.find(r) == -1, lines)
        if newlines != lines:
            open(self.cvs_passfile, "w").writelines(newlines)
    
    def _is_logged_in(self):
        """
        check whether we are still logged into cvs
        """
        if open(self.cvs_passfile, "r").read().find(self.repository) != -1:
            return 1
        return 0

    def _gen_cmd(self, command, options="", cmdoptions=""):
        """
        Generate a cvs command string
        """
        if command not in _cvs_commands:
            raise CVSError, "%s is not a supported cvs command" % command

        cmd = ""

        # first add environment variables
        for pair in self.environment.items():
            cmd = cmd + '%s="%s" ' % pair
        
        cmd = cmd + "%s -d %s " % (self.cvs_bin, self.repository)
        if self.comp_level != "" and self.comp_level != "0":
            cmd = cmd + "-z%s " % self.comp_level
        if options != "":
            cmd = cmd + options + " "
        cmd = cmd + command
        if cmdoptions != "":
            cmd = cmd + " " + cmdoptions
        return cmd
        
    def _exec_cmd(self, command, raise_error=1):
        """
        Execute a command and process the streams
        """
        shellproc = ShellCommandProcessor()
        result_val, result_out, result_err = shellproc.processStreams(command)
        if self.debug:
            # debug mode simply means write result_err to sys.stderr
            try:
                sys.stderr.write(result_err)
            except IOError:
                # ouch - stderr is not writable
                pass
        if result_val and raise_error:
            #XXX this is not very portable, but very efficient
            self._clean_up()
            raise CVSError, result_err
        return result_val, result_out, result_err

    def _prepare(self):
        # change into the module directory
        self.olddir = os.getcwd()
        os.chdir(os.path.join(self.work_dir, self.module))

    def _clean_up(self):
        try:
            olddir = self.olddir
        except AttributeError:
            return
        # change back to the old directory
        os.chdir(olddir)
        del self.olddir

    ##########################################################################
    # Implementation methods for interface
    # IRevisionControl.IRevisionController
    #
    # standard CVS actions

    def init_repository(self):
        """
        initialize CVS repository
        only available if access method "local" has been chosen
        """
        if self.access_method != "local":
            raise CVSError, "Can only initialize local repositories"
        if not os.path.isdir(self.repository_path):
            os.mkdir(self.repository_path)
        command = self._gen_cmd("init")

        result_val, result_out, result_err = self._exec_cmd(command)

    def status(self, filename=""):
        """
        invoke CVS status and parse the output
        don't use the -v option 
        """
        self._prepare()
        command = self._gen_cmd("status", cmdoptions=filename)
        result_val, result_out, result_err = self._exec_cmd(command)
        self._clean_up()
        return self.parse_status(result_out)

    def checkout(self, filename=""):
        """
        perform a CVS checkout
        """
        # change into working directory
        olddir = os.getcwd()
        os.chdir(self.work_dir)

        opt = "%s %s" % (self.module, filename)
        command = self._gen_cmd("checkout", cmdoptions=opt)
        result_val, result_out, result_err = self._exec_cmd(command)
        os.chdir(olddir)

        return self.parse_listout(result_out)

    def commit(self, filename="", log_message="Checkin by CVSWrapper", opt_param=""):
        """
        perform a CVS checkin/commit
        """
        self._prepare()
        opt = '-m "%s" %s %s' % (log_message, opt_param, filename)
        command = self._gen_cmd("commit", cmdoptions=opt)
        result_val, result_out, result_err = self._exec_cmd(command)
        self._clean_up()
        return self.parse_listout(result_out)

    checkin = commit

    def add(self, filename, binary=0):
        """
        add a file to the repository
        """
        self._prepare()
        if binary:
            opt = "-kb " + filename
        else:
            opt = filename
        command = self._gen_cmd("add", cmdoptions=opt)
        result_val, result_out, result_err = self._exec_cmd(command)
        self._clean_up()
        return result_out

    def remove(self, filename):
        """
        remove a file from the repository
        note: the actual file on the filesystem must NOT exist anymore
        """
        self._prepare()
        command = self._gen_cmd("remove", cmdoptions=filename)
        result_val, result_out, result_err = self._exec_cmd(command)
        self._clean_up()
        return result_out

    def update(self, filename="", opt_param="", opt_tail=""):
        """
        perform a CVS update
        """
        self._prepare()
        opt = "%s %s %s" % (opt_param, filename, opt_tail)
        command = self._gen_cmd("update", cmdoptions=opt)
        result_val, result_out, result_err = self._exec_cmd(command)
        self._clean_up()
        return self.parse_listout(result_out)

    def import_module(self, import_dir, vendor_tag, release_tag,
		      log_message="CVS import", ignore_files=[]):
        """
        Do an initial import into the cvs repository
        """
        # change into working directory
        olddir = os.getcwd()
        os.chdir(import_dir)

        opt = '-m "%s" ' % log_message
        for file in ignore_files:
           opt += "-I %s " % file 
        opt += "%s %s %s" % (self.module, vendor_tag, release_tag)
        
        command = self._gen_cmd("import", cmdoptions=opt)
        result_val, result_out, result_err = self._exec_cmd(command)
        os.chdir(olddir)
        
        return self.parse_listout(result_out)

    def tag(self, tag, filename=""):
        """
        tag file(s)
        """
        self._prepare()
        opt = '%s %s' % (tag, filename)
        command = self._gen_cmd("tag", cmdoptions=opt)
        result_val, result_out, result_err = self._exec_cmd(command)
        self._clean_up()
        return result_out

    def log(self, filename=""):
        """
        return log
        """
        self._prepare()
        command = self._gen_cmd("log", cmdoptions=filename)
        result_val, result_out, result_err = self._exec_cmd(command)
        self._clean_up()
        return self.parse_log(result_out)

    def diff(self, filename, rev1="", rev2=""):
        """
        Return a unified diff difference of a file between revision rev1 and revision rev2
        """
        if rev1 != "":
            rev1 = "-r" + rev1
        if rev2 != "":
            rev2 = "-r" + rev2

        self._prepare()
        #opt = "%s %s -u %s" % (rev1, rev2, filename)
        opt = "%s %s %s" % (rev1, rev2, filename)
        command = self._gen_cmd("diff", cmdoptions=opt) 
        # cvs diff returns a positive return value if the revisions differ
        # thus, do not raise an error now
        result_val, result_out, result_err = self._exec_cmd(command, 0)
        self._clean_up()
        
        # instead, raise an error if there is no diff but the return value is
        # still positive
        if not len(result_out) and result_val:
            raise CVSError, result_err
        return result_out
    
    #
    ##########################################################################
    # parsers
    #

    def parse_status(self, output):
        """
        parse the output of the cvs status command
        also support the -v option for displaying tag information
        """
        returndict = {}
        patterns = [ r'File: (?P<filename>.+?)\s',
                     r'.+Status: (?P<status>.+?)\n',
                     r'.+Working revision:\s+(?P<wrev>.+?)\s',
                     r'.+Repository revision:\s+(?P<rrev>.+?)\s+(?P<rfilename>.+?)\n',
                     r'.+Sticky Tag:\s+(?P<stag>.+?)\n',
                     r'.+Sticky Date:\s+(?P<sdate>.+?)\n',
                     r'.+Sticky Options:\s+(?P<sopt>.+?)\n',
                     r'.+\n\s{8}(?P<tagname>.+?)\s+\((?P<taglink>.+?)\)\n'  #XXX doesn't seem to work ?!?
                     ]
        for i in range(len(patterns)):
            patterns[i] = re.compile(patterns[i], re.DOTALL)
        
        # length of the repository and module directory name
        try:
            rname_len = len(os.path.join(self.repository_path, self.module)) + 1
        except IndexError:
            rname_len = 0

        chunks = output.split("===================================================================\n")
        for chunk in chunks:
            chunkdict = {'tags': {}}
            for pattern in patterns:
                match = pattern.match(chunk)
                if match is not None:
                    groupdict = match.groupdict()
                    if groupdict.has_key('tagname'):
                        chunkdict['tags'][groupdict['tagname']] = groupdict['taglink']
                    else:
                        chunkdict.update(match.groupdict())
            relfname = chunkdict.get('rfilename', '')
            if relfname:
                relfname = relfname[rname_len:-2]
                returndict[relfname] = chunkdict
        return returndict
    
    #XXX redo this using regular expressions?!?
    def parse_log(self, log_output):
        """
        parse the output of the cvs log command
        """        
        returndict = {}
        chunks = log_output.split(
	    "=============================================================================\n"
	   )
        for fileinfo in chunks:
            d = {}
            infochunks = fileinfo.split("----------------------------\n")

            lines = infochunks[0].split('\n')
            if len(lines) < 11:
                continue

            lines = map(lambda line: line.split(), lines)
            relfname = lines[2][2]

            # parse main information
            try: d['head'] = lines[3][1]
            except IndexError: d['head'] = None
            try: d['branch'] = lines[4][1]
            except IndexError: d['branch'] = None
            try: d['locks'] = lines[5][1]
            except IndexError: d['locks'] = None
            try: d['access'] = lines[6][2]
            except IndexError: d['access'] = None

            # parse tags
            tags = {}
            for tagline in lines[8:-4]:
                tags[tagline[0][:-1]] = tagline[1]
            d['tags'] = tags

            # parse revisions
            d['revisions'] = []
            for revinfo in infochunks[1:]:
                origlines = revinfo.split('\n')
                lines = map(lambda line: line.split(), origlines)
                rev = {
                    'revision': lines[0][1],
                    'date':     lines[1][1] + " " + lines[1][2][:-1],
                    'author':   lines[1][4][:-1],
                    'state':    lines[1][6][:-1],
                }
                if lines[2][0] == "branches:":
                    rev['log'] = "\n".join(origlines[3:-1])
                    rev['branches'] = origlines[2][11:-1]
                else:
                    rev['log'] = "\n".join(origlines[2:-1])
                    rev['branches'] = None

                # is there a +/- line info?
                if len(lines[1]) > 7:
                    rev['lines'] = lines[1][8] + " " + lines[1][9]
                else:
                    rev['lines'] = None

                d['revisions'].append(rev)
            returndict[relfname] = d
        return returndict

    def parse_listout(self, output, flagmapping=_update_flags):
        """
        parse the output of cvs update/checkout command
        """
        expr = "^([UPARMCINT?]) (.*)$"
        pattern = re.compile(expr)
        lines = output.split('\n')
        returndict = {}
        for line in lines:
            match = pattern.match(line)
            if match is not None:
                flag, filename = match.groups()
                flag = flagmapping.get(flag, flag)
                returndict[filename] = flag

        return returndict
